@@ -66,21 +66,33 @@ matter_arr <- function(data, type = "double", path = NULL,
 		if ( is.null(dimnames) )
 			dimnames <- dimnames(data)
 	}
-	if ( is.null(path) )
-		path <- tempfile(tmpdir=getOption("matter.dump.dir"), fileext=".bin")
-	path <- normalizePath(path, mustWork=FALSE)
-	exists <- file.exists(path)
+	if ( is.null(path) ) {
+		readonly <- FALSE
+		path <- tempfile(tmpdir=getOption("matter.temp.dir"), fileext=".bin")
+		refs <- list(matter_shared_resource(create=path))
+	} else {
+		refs <- NULL
+		if ( any(tolower(path) %in% ":memory:") ) {
+			readonly <- FALSE
+			for ( i in which(tolower(path) %in% ":memory:") ) {
+				path[i] <- tempmem()
+				refs <- c(refs, list(matter_shared_resource(create=path[i])))
+			}
+		}
+	}
+	path <- fix_path(path, mustWork=FALSE)
+	exists <- file.exists(path) | is_shared_memory_object(path)
 	if ( append ) {
 		readonly <- FALSE
-		eof <- file.size(path)
+		eof <- sizeof_shared_resource(path)
 		offset <- ifelse(exists, offset + eof, offset)
 	}
 	if ( is.na(readonly) )
 		readonly <- all(exists) && !missing(data) && !is.null(data)
 	if ( any(exists) && !readonly && !missing(data) && !is.null(data) ) {
-		overwrite <- offset < file.size(path)
+		overwrite <- offset < file.size(path) & !is_shared_memory_pattern(path)
 		if ( any(overwrite) )
-			warning("data may overwrite existing file(s): ",
+			matter_warn("data may overwrite existing file(s): ",
 				paste0(sQuote(path[overwrite]), collapse=", "))
 	}
 	if ( all(exists) && missing(data) ) {
@@ -101,17 +113,18 @@ matter_arr <- function(data, type = "double", path = NULL,
 		sizes <- sizeof(type) * extent
 		offset <- cumsum(c(offset, sizes[-length(sizes)]))
 	}
-	if ( any(!exists) ) {
+	files <- !is_shared_memory_pattern(path)
+	newfiles <- !exists & files
+	if ( any(newfiles) ) {
 		if ( missing(data) && any(extent > 0) && !is.null(data) )
-			warning("creating uninitialized backing file(s): ",
-				paste0(sQuote(path[!exists]), collapse=", "))
-		newfile <- path[!exists]
-		success <- file.create(newfile)
+			matter_warn("creating uninitialized backing file(s): ",
+				paste0(sQuote(path[newfiles]), collapse=", "))
+		success <- file.create(path[newfiles])
 		if ( !all(success) )
-			stop("error creating file(s): ",
-				paste0(sQuote(newfile[!success]), collapse=", "))
+			matter_error("error creating file(s): ",
+				paste0(sQuote(path[newfiles][!success]), collapse=", "))
 	}
-	path <- normalizePath(path, mustWork=TRUE)
+	path[files] <- fix_path(path[files], mustWork=TRUE)
 	x <- new("matter_arr",
 		data=atoms(
 			source=path,
@@ -119,7 +132,8 @@ matter_arr <- function(data, type = "double", path = NULL,
 			offset=as.double(offset),
 			extent=as.double(extent),
 			group=0L,
-			readonly=readonly),
+			readonly=readonly,
+			refs=refs),
 		type=topmode_Rtype(type),
 		dim=dim,
 		dimnames=dimnames,
@@ -129,8 +143,12 @@ matter_arr <- function(data, type = "double", path = NULL,
 		x <- as(x, "matter_vec")
 	if ( length(dim) == 2L )
 		x <- as(x, "matter_mat")
-	if ( !missing(data) && !is.null(data) )
-		x[] <- data
+	if ( !missing(data) ) {
+		if ( any(is_shared_memory_pattern(path)) )
+			requisition_atoms(atomdata(x))
+		if ( !is.null(data) )
+			x[] <- data
+	}
 	x
 }
 
@@ -178,7 +196,7 @@ matter_mat <- function(data, type = "double", path = NULL,
 		if ( is.na(nrow) && !rowMaj )
 			nrow <- max(extent)
 	}
-	x <- matter_arr(data=NULL, type=type, path=path, dim=c(nrow, ncol),
+	x <- matter_arr(NULL, type=type, path=path, dim=c(nrow, ncol),
 		dimnames=dimnames, offset=offset, extent=extent,
 		readonly=readonly, append=append, rowMaj=rowMaj, ...)
 	x <- as(x, "matter_mat")
@@ -189,8 +207,9 @@ matter_mat <- function(data, type = "double", path = NULL,
 		n1 <- ncol(x) # number of atoms/groups
 		n2 <- nrow(x) # extent of each atom/group
 	}
+	extent <- as.numeric(atomdata(x)@extent)
 	if ( length(x@data) == n1 && all(unique(extent) == n2) ) {
-		x@data <- regroup_atoms(x@data, 0L:(n1 - 1L))
+		x@data <- setgroup_atoms(x@data, 0L:(n1 - 1L))
 	} else {
 		x@data <- regroup_atoms(ungroup_atoms(x@data), n1)
 	}
@@ -251,41 +270,49 @@ setMethod("as.double", "matter_arr",
 setMethod("as.numeric", "matter_arr",
 	function(x, ...) as.vector(x, "double"))
 
-setMethod("as.matrix", "matter_arr",
-	function(x, ...) {
-		y <- as(x, "matter_mat")
+setAs("matter_arr", "matrix",
+	function(from) {
+		to <- as(from, "matter_mat")
 		if ( getOption("matter.coerce.altrep") ) {
-			as.altrep(y)
+			as.altrep(to)
 		} else {
-			y[]
+			to[]
 		}
 	})
 
-setMethod("as.array", "matter_arr",
-	function(x, ...) {
-		y <- as(x, "matter_arr")
+setAs("matter_arr", "array",
+	function(from) {
+		to <- as(from, "matter_arr")
 		if ( getOption("matter.coerce.altrep") ) {
-			as.altrep(y)
+			as.altrep(to)
 		} else {
-			y[]
+			to[]
 		}
 	})
 
+as.matrix.matter_arr <- function(x, ...) as(x, "matrix")
+
+as.array.matter_arr <- function(x, ...) as(x, "array")
+
+setMethod("as.matrix", "matter_arr", as.matrix.matter_arr)
+
+setMethod("as.array", "matter_arr", as.array.matter_arr)
+
 setMethod("describe_for_display", "matter_arr", function(x) {
 	desc1 <- paste0("<", paste0(dim(x), collapse=" x "), " dim> ", class(x))
-	desc2 <- paste0("out-of-memory ", type(x), " array")
+	desc2 <- paste0("out-of-core ", type(x), " array")
 	paste0(desc1, " :: ", desc2)
 })
 
 setMethod("describe_for_display", "matter_mat", function(x) {
 	desc1 <- paste0("<", nrow(x), " row x ", ncol(x), " col> ", class(x))
-	desc2 <- paste0("out-of-memory ", type(x), " matrix")
+	desc2 <- paste0("out-of-core ", type(x), " matrix")
 	paste0(desc1, " :: ", desc2)
 })
 
 setMethod("describe_for_display", "matter_vec", function(x) {
 	desc1 <- paste0("<", length(x), " length> ", class(x))
-	desc2 <- paste0("out-of-memory ", type(x), " vector")
+	desc2 <- paste0("out-of-core ", type(x), " vector")
 	paste0(desc1, " :: ", desc2)
 })
 
@@ -300,18 +327,191 @@ setMethod("preview_for_display", "matter_arr", function(x) {
 	}
 })
 
+setMethod("mem_realized", "matter_arr", function(x) {
+	size_bytes(sum(length(x) * sizeof(type(x)), na.rm=TRUE))
+})
+
+copy_to_matter_fun <- function(id, dest, margin = NULL)
+{
+	if ( !is.null(margin) && !margin %in% c(1L, 2L) )
+		matter_error("'margin' must be 1 or 2")
+	isoclos(function(src) {
+		i <- attr(src, "index")
+		ipclock(id)
+		if ( is.null(margin) ) {
+			dest[i] <- src
+		} else {
+			switch(margin,
+				dest[i,] <- src,
+				dest[,i] <- src)
+		}
+		ipcunlock(id)
+	}, matter_env())
+}
+
+copy_to_matter_arr <- function(object, type = NULL, path = NULL,
+	offset = 0, readonly = TRUE, append = FALSE,
+	rowMaj = FALSE, ..., BPPARAM)
+{
+	BPPARAM <- bplocalized(BPPARAM)
+	if ( is.matter(object) ) {
+		type <- type %||% type(atomdata(object))
+		extent <- lengths(atomdata(object))
+	} else {
+		type <- type %||% type(object)
+		extent <- length(object)
+	}
+	x <- matter_arr(NULL, type=type, path=path,
+		dim=dim(object), dimnames=dimnames(object),
+		offset=offset, extent=extent, readonly=FALSE,
+		append=append, rowMaj=rowMaj)
+	type(x) <- type(object)
+	pid <- ipcid()
+	FUN <- copy_to_matter_fun(pid, x)
+	chunk_lapply(object, FUN, ..., BPPARAM=BPPARAM)
+	readonly(x) <- readonly
+	ipcremove(pid)
+	if ( validObject(x) )
+		x
+}
+
+copy_to_matter_mat <- function(object, type = NULL, path = NULL,
+	offset = 0, readonly = TRUE, append = FALSE,
+	rowMaj = FALSE, ..., BPPARAM)
+{
+	BPPARAM <- bplocalized(BPPARAM)
+	if ( is.matter(object) ) {
+		type <- type %||% type(atomdata(object))
+		extent <- lengths(atomdata(object))
+	} else {
+		type <- type %||% type(object)
+		extent <- length(object)
+	}
+	x <- matter_mat(NULL, type=type, path=path,
+		nrow=nrow(object), ncol=ncol(object), dimnames=dimnames(object),
+		offset=offset, extent=extent, readonly=FALSE,
+		append=append, rowMaj=rowMaj)
+	type(x) <- type(object)
+	pid <- ipcid()
+	if ( rowMaj ) {
+		FUN <- copy_to_matter_fun(pid, x, 1L)
+		chunk_rowapply(object, FUN, ..., BPPARAM=BPPARAM)
+	} else {
+		FUN <- copy_to_matter_fun(pid, x, 2L)
+		chunk_colapply(object, FUN, ..., BPPARAM=BPPARAM)
+	}
+	readonly(x) <- readonly
+	ipcremove(pid)
+	if ( validObject(x) )
+		x
+}
+
+copy_to_matter_vec <- function(object, type = NULL, path = NULL,
+	offset = 0, readonly = TRUE, append = FALSE,
+	rowMaj = FALSE, ..., BPPARAM)
+{
+	BPPARAM <- bplocalized(BPPARAM)
+	if ( is.matter(object) ) {
+		type <- type %||% type(atomdata(object))
+		extent <- lengths(atomdata(object))
+	} else {
+		type <- type %||% type(object)
+		extent <- length(object)
+	}
+	x <- matter_vec(NULL, type=type, path=path,
+		length=length(object), names=names(object),
+		offset=offset, extent=extent, readonly=FALSE,
+		append=append, rowMaj=rowMaj)
+	type(x) <- type(object)
+	pid <- ipcid()
+	FUN <- copy_to_matter_fun(pid, x)
+	chunk_lapply(object, FUN, ..., BPPARAM=BPPARAM)
+	readonly(x) <- readonly
+	ipcremove(pid)
+	if ( validObject(x) )
+		x
+}
+
+setMethod("fetch", "matter_arr",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_arr(object, path=":memory:", ..., BPPARAM=BPPARAM))
+
+setMethod("flash", "matter_arr",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_arr(object, ..., BPPARAM=BPPARAM))
+
+setMethod("fetch", "matter_mat",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_mat(object, path=":memory:", ..., BPPARAM=BPPARAM))
+
+setMethod("flash", "matter_mat",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_mat(object, ..., BPPARAM=BPPARAM))
+
+setMethod("fetch", "matter_vec",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_vec(object, path=":memory:", ..., BPPARAM=BPPARAM))
+
+setMethod("flash", "matter_vec",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_vec(object, ..., BPPARAM=BPPARAM))
+
+setMethod("fetch", "raw",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_vec(object, path=":memory:", ..., BPPARAM=BPPARAM))
+
+setMethod("flash", "raw",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_vec(object, ..., BPPARAM=BPPARAM))
+
+setMethod("fetch", "logical",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_vec(object, path=":memory:", ..., BPPARAM=BPPARAM))
+
+setMethod("flash", "logical",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_vec(object, ..., BPPARAM=BPPARAM))
+
+setMethod("fetch", "numeric",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_vec(object, path=":memory:", ..., BPPARAM=BPPARAM))
+
+setMethod("flash", "numeric",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_vec(object, ..., BPPARAM=BPPARAM))
+
+setMethod("fetch", "matrix",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_mat(object, path=":memory:", ..., BPPARAM=BPPARAM))
+
+setMethod("flash", "matrix",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_mat(object, ..., BPPARAM=BPPARAM))
+
+setMethod("fetch", "array",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_arr(object, path=":memory:", ..., BPPARAM=BPPARAM))
+
+setMethod("flash", "array",
+	function(object, ..., BPPARAM = bpparam())
+		copy_to_matter_arr(object, ..., BPPARAM=BPPARAM))
+
 subset_matter_arr_elts <- function(x, i = NULL)
 {
 	if ( x@transpose ) {
-		index <- array_ind(i, dim(x))
-		i <- linear_ind(index, dim(x), rowMaj=TRUE)
+		index <- array_ind(i, x@dim)
+		i <- linear_ind(index, x@dim, rowMaj=TRUE)
 	}
 	data <- ungroup_atoms(x@data)
 	data <- subset_atoms2(data, i, NULL)
-	if ( !is.null(x@ops) )
-		warning("deferred operations will be dropped")
+	if ( !is.null(dim(x)) && !is.null(x@ops) ) {
+		ops <- NULL
+		matter_warn("deferred operations will be dropped")
+	} else {
+		ops <- subset_ops(x@ops, list(i))
+	}
 	y <- new(class(x), x, data=data, dim=length(i),
-		names=x@names[i], dimnames=NULL, ops=NULL)
+		names=x@names[i], dimnames=NULL, ops=ops)
 	if ( validObject(y) )
 		y
 }
@@ -340,10 +540,9 @@ subset_matter_arr_subarray <- function(x, index)
 		if (is.null(index[[j]])) dim(x)[j]
 		else length(index[[j]]), numeric(1))
 	dnm <- subset_dimnames(x@dimnames, index)
-	if ( !is.null(x@ops) )
-		warning("deferred operations will be dropped")
+	ops <- subset_ops(x@ops, index)
 	y <- new(class(x), x, data=data, dim=dm,
-		dimnames=dnm, ops=NULL)
+		dimnames=dnm, ops=ops)
 	if ( validObject(y) )
 		y
 }
@@ -382,10 +581,9 @@ subset_matter_mat_submatrix <- function(x, i = NULL, j = NULL)
 	ncol <- if (is.null(j)) x@dim[2L] else length(j)
 	dm <- c(nrow, ncol)
 	dnm <- subset_dimnames(x@dimnames, list(i, j))
-	if ( !is.null(x@ops) )
-		warning("deferred operations will be dropped")
+	ops <- subset_ops(x@ops, list(i, j))
 	y <- new(class(x), x, data=data, dim=dm,
-		dimnames=dnm, ops=NULL)
+		dimnames=dnm, ops=ops)
 	if ( validObject(y) )
 		y
 }
@@ -409,14 +607,14 @@ setMethod("[", c(x = "matter_arr"),
 		narg <- nargs() - 1L - !missing(drop)
 		if ( (narg == 1L && !missing(i)) || is.null(dim(x)) ) {
 			i <- as_subscripts(i, x)
-			if ( is_nil(drop) ) {
+			if ( is_null_or_na(drop) ) {
 				subset_matter_arr_elts(x, i)
 			} else {
 				get_matter_arr_elts(x, i)
 			}
 		} else {
 			if ( narg != 1L && narg != length(dim(x)) )
-				stop("incorrect number of dimensions")
+				matter_error("incorrect number of dimensions")
 			if ( missing(i) ) i <- NULL
 			if ( missing(j) ) j <- NULL
 			if ( ...length() > 0 ) {
@@ -431,7 +629,9 @@ setMethod("[", c(x = "matter_arr"),
 			} else {
 				index <- list(i)
 			}
-			if ( is_nil(drop) ) {
+			if ( !any(non_null(index)) && missing(drop) )
+				drop <- FALSE
+			if ( is_null_or_na(drop) ) {
 				subset_matter_arr_subarray(x, index)
 			} else {
 				get_matter_arr_subarray(x, index, drop)
@@ -447,7 +647,7 @@ setReplaceMethod("[", c(x = "matter_arr"),
 			set_matter_arr_elts(x, i, value)
 		} else {
 			if ( narg != 1L && narg != length(dim(x)) )
-				stop("incorrect number of dimensions")
+				matter_error("incorrect number of dimensions")
 			if ( missing(i) ) i <- NULL
 			if ( missing(j) ) j <- NULL
 			if ( ...length() > 0 ) {
@@ -474,17 +674,19 @@ setMethod("[", c(x = "matter_mat"),
 			get_matter_arr_elts(x, i)
 		} else {
 			if ( narg != 1L && narg != 2L )
-				stop("incorrect number of dimensions")
+				matter_error("incorrect number of dimensions")
+			if ( missing(i) && missing(j) && missing(drop) )
+				drop <- FALSE
 			i <- as_row_subscripts(i, x)
 			j <- as_col_subscripts(j, x)
 			if ( isTRUE(x@indexed) ) {
-				if ( is_nil(drop) ) {
+				if ( is_null_or_na(drop) ) {
 					subset_matter_mat_submatrix(x, i, j)
 				} else {
 					get_matter_mat_submatrix(x, i, j, drop)
 				}
 			} else {
-				if ( is_nil(drop) ) {
+				if ( is_null_or_na(drop) ) {
 					subset_matter_arr_subarray(x, list(i, j))
 				} else {
 					get_matter_arr_subarray(x, list(i, j), drop)
@@ -501,7 +703,7 @@ setReplaceMethod("[", c(x = "matter_mat"),
 			set_matter_arr_elts(x, i, value)
 		} else {
 			if ( narg != 1L && narg != 2L )
-				stop("incorrect number of dimensions")
+				matter_error("incorrect number of dimensions")
 			i <- as_row_subscripts(i, x)
 			j <- as_col_subscripts(j, x)
 			if ( isTRUE(x@indexed) ) {
@@ -515,12 +717,12 @@ setReplaceMethod("[", c(x = "matter_mat"),
 setMethod("combine", "matter_arr",
 	function(x, y, ...) {
 		if ( rowMaj(x) || rowMaj(y) )
-			stop("can't combine row-major arrays")
+			matter_error("can't combine row-major arrays")
 		data <- rbind(
 			ungroup_atoms(x@data),
 			ungroup_atoms(y@data))
 		if ( !is.null(x@ops) || !is.null(y@ops) )
-			warning("deferred operations will be dropped")
+			matter_warn("deferred operations will be dropped")
 		new("matter_vec", data=data, type=x@type,
 			dim=length(x) + length(y),
 			names=combine_names(x, y),
@@ -543,11 +745,11 @@ setMethod("cbind2", c("matter_vec", "matter_mat"),
 setMethod("cbind2", c("matter_mat", "matter_mat"),
 	function(x, y, ...) {
 		if ( nrow(x) != nrow(y) )
-			stop("number of rows of matrices must match")
+			matter_error("number of rows of matrices must match")
 		if ( rowMaj(x) || rowMaj(y) )
-			stop("can't cbind row-major matrices")
+			matter_error("can't cbind row-major matrices")
 		if ( !is.null(x@ops) || !is.null(y@ops) )
-			warning("deferred operations will be dropped")
+			matter_warn("deferred operations will be dropped")
 		data <- cbind(x@data, y@data)
 		new(class(x), x, data=data,
 			dim=c(nrow(x), ncol(x) + ncol(y)),
@@ -567,11 +769,11 @@ setMethod("rbind2", c("matter_vec", "matter_mat"),
 setMethod("rbind2", c("matter_mat", "matter_mat"),
 	function(x, y, ...) {
 		if ( ncol(x) != ncol(y) )
-			stop("number of columns of matrices must match")
+			matter_error("number of columns of matrices must match")
 		if ( !x@transpose || !y@transpose )
-			stop("can't rbind column-major matrices")
+			matter_error("can't rbind column-major matrices")
 		if ( !is.null(x@ops) || !is.null(y@ops) )
-			warning("deferred operations will be dropped")
+			matter_warn("deferred operations will be dropped")
 		data <- cbind(x@data, y@data)
 		new(class(x), x, data=data,
 			dim=c(nrow(x) + nrow(y), ncol(x)),
@@ -617,6 +819,12 @@ setMethod("Arith", c(e1 = "vector", e2 = "matter_arr"),
 setMethod("Arith", c(e1 = "array", e2 = "matter_arr"),
 	function(e1, e2) register_op(e2, .Generic, e1, TRUE))
 
+setMethod("exp", "matter_arr", function(x) register_op(x, "exp"))
+setMethod("log", "matter_arr", function(x) register_op(x, "log"))
+setMethod("log2", "matter_arr", function(x) register_op(x, "log2"))
+setMethod("log10", "matter_arr", function(x) register_op(x, "log10"))
+setMethod("log1p", "matter_arr", function(x) register_op(x, "log1p"))
+
 setMethod("rowMaj", "matter_arr", function(x)
 {
 	isTRUE(x@transpose)
@@ -624,9 +832,12 @@ setMethod("rowMaj", "matter_arr", function(x)
 
 setMethod("t", "matter_arr", function(x)
 {
+	if ( length(dim(x)) > 2L )
+		matter_error("argument is not a matrix")
 	x@transpose <- !x@transpose
 	x@dim <- rev(x@dim)
 	x@dimnames <- rev(x@dimnames)
+	x@ops <- t_ops(x@ops)
 	if ( validObject(x) )
 		x
 })
@@ -708,21 +919,16 @@ lmatmul <- function(x, y, useOuter = FALSE)
 rmatmul_sc <- function(x, y, useOuter = FALSE)
 {
 	nchunks <- getOption("matter.default.nchunks")
-	verbose <- getOption("matter.default.verbose")
 	ans <- matrix(0, nrow=nrow(x), ncol=ncol(y))
 	if ( useOuter ) {
 		INDEX <- chunkify(seq_len(ncol(x)), nchunks)
 		for ( i in INDEX ) {
-			if ( verbose )
-				print_chunk_progress(i, length(INDEX))
 			xi <- as.matrix(x[,i,drop=FALSE])
 			ans <- ans + xi %*% y[i,,drop=FALSE]
 		}
 	} else {
 		INDEX <- chunkify(seq_len(nrow(x)), nchunks)
 		for ( i in INDEX ) {
-			if ( verbose )
-				print_chunk_progress(i, length(INDEX))
 			xi <- as.matrix(x[i,,drop=FALSE])
 			ans[i,] <- xi %*% y
 		}
@@ -735,15 +941,15 @@ rmatmul_mc <- function(x, y, useOuter = FALSE, BPPARAM = NULL)
 {
 	if ( useOuter ) {
 		add <- function(...) Reduce("+", list(...))
-		matmul <- function(xi) {
-			i <- attr(xi, "index")
-			xi %*% y[i,,drop=FALSE]
-		}
-		ans <- chunk_colapply(x, matmul,
+		matmul <- isofun(function(xi, y) {
+				i <- attr(xi, "index")
+				xi %*% y[i,,drop=FALSE]
+			})
+		ans <- chunk_colapply(x, matmul, y=y,
 			simplify=add, BPPARAM=BPPARAM)
 	} else {
-		matmul <- function(xi) xi %*% y
-		ans <- chunk_rowapply(x, matmul,
+		matmul <- isofun(function(xi, y) xi %*% y)
+		ans <- chunk_rowapply(x, matmul, y=y,
 			simplify=rbind, BPPARAM=BPPARAM)
 	}
 	ans
@@ -753,21 +959,16 @@ rmatmul_mc <- function(x, y, useOuter = FALSE, BPPARAM = NULL)
 lmatmul_sc <- function(x, y, useOuter = FALSE)
 {
 	nchunks <- getOption("matter.default.nchunks")
-	verbose <- getOption("matter.default.verbose")
 	ans <- matrix(0, nrow=nrow(x), ncol=ncol(y))
 	if ( useOuter ) {
 		INDEX <- chunkify(seq_len(nrow(y)), nchunks)
 		for ( i in INDEX ) {
-			if ( verbose )
-				print_chunk_progress(i, length(INDEX))
 			yi <- as.matrix(y[i,,drop=FALSE])
 			ans <- ans + x[,i,drop=FALSE] %*% yi
 		}
 	} else {
 		INDEX <- chunkify(seq_len(ncol(y)), nchunks)
 		for ( i in INDEX ) {
-			if ( verbose )
-				print_chunk_progress(i, length(INDEX))
 			yi <- as.matrix(y[,i,drop=FALSE])
 			ans[,i] <- x %*% yi
 		}
@@ -780,15 +981,15 @@ lmatmul_mc <- function(x, y, useOuter = FALSE, BPPARAM = NULL)
 {
 	if ( useOuter ) {
 		add <- function(...) Reduce("+", list(...))
-		matmul <- function(yi) {
-			i <- attr(yi, "index")
-			x[,i,drop=FALSE] %*% yi
-		}
-		ans <- chunk_rowapply(y, matmul,
+		matmul <- isofun(function(yi, x) {
+				i <- attr(yi, "index")
+				x[,i,drop=FALSE] %*% yi
+			})
+		ans <- chunk_rowapply(y, matmul, x=x,
 			simplify=add, BPPARAM=BPPARAM)
 	} else {
-		matmul <- function(yi) x %*% yi
-		ans <- chunk_colapply(y, matmul,
+		matmul <- isofun(function(yi, x) x %*% yi)
+		ans <- chunk_colapply(y, matmul, x=x,
 			simplify=cbind, BPPARAM=BPPARAM)
 	}
 	ans
