@@ -11,6 +11,11 @@
 // (avoid duplicates or division by 0)
 #define EXACT 0
 
+// search direction
+#define NODE_PARENT	0
+#define NODE_LEFT	1
+#define NODE_RIGHT	2
+
 // swap items (use with caution)
 #define swap(x, y, T) do { T swap = x; x = y; y = swap; } while (false)
 
@@ -251,7 +256,7 @@ void do_quick_sort(int * ptr, T * x, size_t start, size_t end, bool ind1 = false
 	T * dup = R_Calloc(n, T);
 	std::memcpy(dup, x + start, n * sizeof(T));
 	quick_sort(dup, 0, n, ptr);
-	Free(dup);
+	R_Free(dup);
 }
 
 // sort an array x and return ranks in ptr
@@ -296,8 +301,8 @@ index_t do_quick_rank(int * ptr, T * x, size_t start, size_t end, bool ties_max
 			rank += count;
 		}
 	}
-	Free(indx);
-	Free(dup);
+	R_Free(indx);
+	R_Free(dup);
 	return rank;
 }
 
@@ -340,7 +345,7 @@ void do_quick_select(T * ptr, T * x, size_t start, size_t end, int * k, size_t n
 		else 
 			ptr[i] = ptr[i - 1];
 	}
-	Free(dup);
+	R_Free(dup);
 }
 
 //// Median
@@ -369,7 +374,7 @@ double quick_median(T * x, size_t n)
 	}
 	else
 		result = quick_select(dup, 0, n, k);
-	Free(dup);
+	R_Free(dup);
 	return result;
 }
 
@@ -389,7 +394,7 @@ double quick_mad(T * x, size_t n, double center = NA_REAL, double scale = 1.4826
 			dev[i] = std::fabs(x[i] - center);
 	}
 	double mad = scale * quick_median(dev, n);
-	Free(dev);
+	R_Free(dev);
 	return mad;
 }
 
@@ -549,8 +554,8 @@ index_t kd_tree_build(T * x, size_t k, size_t n,
 				xs[i] = x[n * jnext + indx[i]];
 		}
 	}
-	Free(indx);
-	Free(xs);
+	R_Free(indx);
+	R_Free(xs);
 	return root;
 }
 
@@ -614,7 +619,7 @@ index_t kd_tree_search(int * ptr, T * x, T * data, size_t k, size_t n,
 //// K-NN search
 //-----------------
 
-// search for knn points nearest x, return via ptr
+// search for points nearest query x, return via ptr
 template<typename T>
 void knn_search(int * ptr, T * x, T * data, size_t k, size_t n,
 	int * left_child, int * right_child, size_t root, int knn,
@@ -644,19 +649,29 @@ void knn_search(int * ptr, T * x, T * data, size_t k, size_t n,
 		index_t j = depth % k;
 		double ds = sdiff(x[j], data[n * j + node]);
 		double du = std::fabs(ds);
-		double d2 = do_dist(x, data + node, k, 1, n, metric, p);
+		double D = do_dist(x, data + node, k, 1, n, metric, p);
 		// check if this is a better neighbor
-		if ( d2 < best[knn - 1] )
+		if ( D <= best[knn - 1] )
 		{
 			index_t i = knn - 1;
-			ptr[i] = node + ind1;
-			best[i] = d2;
-			// sort this neighbor into place
-			while ( i > 0 && lt(best[i], best[i - 1]) )
+			// process strictly better neighbors _or_ ties
+			if ( D < best[i] || (node + ind1) < ptr[i] )
 			{
-				swap(ptr[i], ptr[i - 1], int);
-				swap(best[i], best[i - 1], double);
-				i--;
+				ptr[i] = node + ind1;
+				best[i] = D;
+				// sort this neighbor into place
+				while ( i > 0 && lteq(best[i], best[i - 1]) )
+				{
+					// ties are broken by index order
+					if ( lt(best[i], best[i - 1]) || ptr[i] < ptr[i - 1] )
+					{
+						swap(ptr[i], ptr[i - 1], int);
+						swap(best[i], best[i - 1], double);
+						i--;
+					}
+					else
+						break;
+				}
 			}
 		}
 		// check if we need to search left subtree
@@ -674,7 +689,7 @@ void knn_search(int * ptr, T * x, T * data, size_t k, size_t n,
 	}
 }
 
-// search for knn points nearest x, return via ptr
+// search for points nearest query x, return via ptr
 template<typename T>
 void do_knn_search(int * ptr, T * x, T * data, size_t k, size_t nx, size_t ndata,
 	int * left_child, int * right_child, size_t root, int knn,
@@ -693,4 +708,138 @@ void do_knn_search(int * ptr, T * x, T * data, size_t k, size_t nx, size_t ndata
 	}
 }
 
+// search for points nearest node x, return via ptr
+template<typename T>
+void knn_self_search(int * ptr, int x, T * data, size_t k, size_t n,
+	int * left_child, int * right_child, int * parent, int knn,
+	int metric = DIST_EUC, double p = 2, bool ind1 = false)
+{
+	if ( n == 0 || k == 0 || knn == 0 )
+		return;
+	index_t node = x, depth = -1, origin = -1;
+	// initialize knn
+	double best [knn];
+	for ( index_t i = 0; i < knn; i++ )
+	{
+		ptr[i] = NA_INTEGER;
+		best[i] = R_PosInf;
+	}
+	// initialize depth
+	while ( !isNA(node) )
+	{
+		node = parent[node];
+		++depth;
+	}
+	// initialize stack (2x size due to traversal up _and_ down)
+	int stack_size = 2 * 3 * std::ceil(std::log2(n) + 1);
+	int stack [stack_size];
+	int top = -1;
+	stack[++top] = x;		// node
+	stack[++top] = depth;	// depth
+	stack[++top] = origin;	// direction
+	while ( top >= 0 )
+	{
+		// pop node
+		origin = stack[top--];
+		depth = stack[top--];
+		node = stack[top--];
+		index_t j = depth % k;
+		double ds = sdiff(data[n * j + x], data[n * j + node]);
+		double du = std::fabs(ds);
+		double D = do_dist(data + x, data + node, k, n, n, metric, p);
+		// check if this is a better neighbor
+		if ( D <= best[knn - 1] )
+		{
+			index_t i = knn - 1;
+			// process strictly better neighbors _or_ ties
+			if ( D < best[i] || (node + ind1) < ptr[i] )
+			{
+				ptr[i] = node + ind1;
+				best[i] = D;
+				// sort this neighbor into place
+				while ( i > 0 && lteq(best[i], best[i - 1]) )
+				{
+					// ties are broken by index order
+					if ( lt(best[i], best[i - 1]) || ptr[i] < ptr[i - 1] )
+					{
+						swap(ptr[i], ptr[i - 1], int);
+						swap(best[i], best[i - 1], double);
+						i--;
+					}
+					else
+						break;
+				}
+			}
+		}
+		// check if we need to search left subtree
+		if ( origin != NODE_LEFT )
+		{
+			if ( (ds < 0 || du <= best[knn - 1]) && !isNA(left_child[node]) )
+			{
+				stack[++top] = left_child[node];
+				stack[++top] = depth + 1;
+				stack[++top] = NODE_PARENT;
+			}
+		}
+		// check if we need to search right subtree
+		if ( origin != NODE_RIGHT )
+		{
+			if ( (ds > 0 || du <= best[knn - 1]) && !isNA(right_child[node]) )
+			{
+				stack[++top] = right_child[node];
+				stack[++top] = depth + 1;
+				stack[++top] = NODE_PARENT;
+			}
+		}
+		// check if we need to search parent
+		if ( origin != NODE_PARENT )
+		{
+			if ( !isNA(parent[node]) )
+			{
+				// check if this is left child
+				if ( node == left_child[parent[node]] )
+				{
+					stack[++top] = parent[node];
+					stack[++top] = depth - 1;
+					stack[++top] = NODE_LEFT;
+				}
+				// check if this is right child
+				if ( node == right_child[parent[node]] )
+				{
+					stack[++top] = parent[node];
+					stack[++top] = depth - 1;
+					stack[++top] = NODE_RIGHT;
+				}
+			}
+		}
+	}
+}
+
+// search for points nearest each node, return via ptr
+template<typename T>
+void do_knn_self_search(int * ptr, T * data, size_t k, size_t n,
+	int * left_child, int * right_child, int knn,
+	int metric = DIST_EUC, double p = 2, bool ind1 = false)
+{
+	int nn [knn];
+	int * parent = R_Calloc(n, int);
+	for ( size_t i = 0; i < n; i++ )
+		parent[i] = NA_INTEGER;
+	for ( size_t i = 0; i < n; i++ )
+	{
+		if ( !isNA(left_child[i]) )
+			parent[left_child[i]] = i;
+		if ( !isNA(right_child[i]) )
+			parent[right_child[i]] = i;
+	}
+	for ( index_t i = 0; i < n; i++ )
+	{
+		knn_self_search(nn, i, data, k, n,
+			left_child, right_child, parent, knn, metric, p, ind1);
+		for ( int l = 0; l < knn; l++ )
+			ptr[l * n + i] = nn[l];
+	}
+	R_Free(parent);
+}
+
 #endif // SEARCH
