@@ -244,7 +244,7 @@ double do_quant(T * x, index_t lower, index_t upper, double prob)
 		k = j - 1;
 	// find the kth order statistic by sorting x
 	T q = quick_select(dup, 0, n, norm_ind(k, n));
-	Free(dup);
+	R_Free(dup);
 	return coerce_cast<double>(q);
 }
 
@@ -367,11 +367,13 @@ void bilateral_filter(T * x, size_t n, int width,
 	int r = width / 2;
 	index_t ij;
 	double sdd = sddist, sdr = sdrange;
-	double mad, xrange;
+	double xmedian, xmad, xrange;
+	double D = r;
 	if ( !isNA(spar) )
 	{
 		// get MAD if using adaptive parameters
-		mad = quick_mad(x, n);
+		xmedian = quick_median(x, n);
+		xmad = quick_mad(x, n);
 		double xmin = do_min(x, 0, n - 1);
 		double xmax = do_max(x, 0, n - 1);
 		xrange = xmax - xmin;
@@ -383,7 +385,7 @@ void bilateral_filter(T * x, size_t n, int width,
 			buffer[i] = NA_REAL;
 			continue;
 		}
-		double dmean = 0, W = 0;
+		double dx = 0, W = 0;
 		buffer[i] = 0;
 		if ( !isNA(spar) )
 		{
@@ -393,12 +395,13 @@ void bilateral_filter(T * x, size_t n, int width,
 				// find mean of local differences
 				ij = norm_ind(i + j, n);
 				if ( !isNA(x[ij]) )
-					dmean += std::fabs(x[ij] - x[i]) / width;
+					dx += std::fabs(x[ij] - xmedian);
 			}
+			dx /= width;
 			// calculate adaptive parameters
-			double z = std::fabs(dmean - mad) / spar;
+			double z = std::fabs(dx - xmad) / spar;
 			if ( isNA(sddist) )
-				sdd = r * std::exp(-z) / std::sqrt(2);
+				sdd = D * std::exp(-z) / std::sqrt(2);
 			if ( isNA(sdrange) )
 				sdr = xrange * std::exp(-z) / std::sqrt(2);
 		}
@@ -480,7 +483,7 @@ void diffusion_filter(T * x, size_t n, int niter,
 			x1[i] = x0[i] + rate * dx;
 		}
 	}
-	Free(tmp);
+	R_Free(tmp);
 }
 
 template<typename T>
@@ -572,8 +575,8 @@ void guided_filter(T * x, T * g, size_t n, int width,
 		else
 			buffer[i] = ua[i] * g[i] + ub[i];
 	}
-	Free(tmp);
-	Free(u);
+	R_Free(tmp);
+	R_Free(u);
 }
 
 //// Warping and alignment
@@ -634,7 +637,7 @@ void warp_dtw(Tx * x, Tx * y, Tt * tx, Tt * ty, int nx, int ny,
 		}
 		k++;
 	}
-	Free(D);
+	R_Free(D);
 }
 
 template<typename Tx, typename Tt>
@@ -737,8 +740,8 @@ void warp_dtwc(Tx * x, Tx * y, Tt * tx, Tt * ty, int nx, int ny,
 		}
 		k++;
 	}
-	Free(ptrs);
-	Free(D);
+	R_Free(ptrs);
+	R_Free(D);
 }
 
 // correlation between x and y (w/ interpolation)
@@ -863,9 +866,9 @@ void warp_cow(Tx * x, Tx * y, Tt * tx, Tt * ty, int nx, int ny,
 		index_t j = (x_nodes[i] - wa[i]);
 		x_nodes[i + 1] = W[pW[i] + j];
 	}
-	Free(ptrs);
-	Free(W);
-	Free(F);
+	R_Free(ptrs);
+	R_Free(W);
+	R_Free(F);
 }
 
 //// Binning and downsampling
@@ -915,8 +918,8 @@ void bin_vector(T * x, int n, int * lower, int * upper,
 			// expand to adjacent bins
 			index_t a = lower[i];
 			index_t b = upper[i];
-			a = a - 1 < 0 ? 0 : a - 1;
-			b = b + 1 > n - 1 ? n - 1 : b + 1;
+			a = norm_ind(a - 1, n);
+			b = norm_ind(b + 1, n);
 			// initialize statistics
 			double ux = do_mean(x, lower[i], upper[i]);
 			double ut = (a + b) / 2;
@@ -1186,7 +1189,7 @@ void smooth_snip(T * x, size_t n, T * buffer, int m, bool decreasing = true)
 				y[i] = z[i];
 		}
 	}
-	Free(z);
+	R_Free(z);
 }
 
 //// Peak detection
@@ -1195,17 +1198,18 @@ void smooth_snip(T * x, size_t n, T * buffer, int m, bool decreasing = true)
 template<typename T>
 size_t local_maxima(T * x, size_t n, int * buffer, int width = 5)
 {
-	int nmax = 0, a = 0, b = n, r = abs(width / 2);
+	int nmax = 0, r = abs(width / 2), a, b;
 	for ( index_t i = 0; i < n; i++ )
 	{
 		buffer[i] = false;
 		if ( i < r || i > n - r )
 			continue;
-		a = (i - r) > 0 ? (i - r) : 0;
-		b = (i + r) < n - 1 ? (i + r) : n - 1;
+		a = norm_ind(i - r, n);
+		b = norm_ind(i + r, n);
 		for ( index_t j = a; j <= b; j++ )
 		{
-			buffer[i] = true;
+			if ( x[i] > x[j] )
+				buffer[i] = true;
 			if ( j < i && x[j] >= x[i] )
 			{
 				buffer[i] = false;
@@ -1216,8 +1220,9 @@ size_t local_maxima(T * x, size_t n, int * buffer, int width = 5)
 				buffer[i] = false;
 				break;
 			}
-			nmax++;
 		}
+		if ( buffer[i] )
+			nmax++;
 	}
 	return nmax;
 }
@@ -1241,7 +1246,7 @@ index_t peak_lbound(T * x, index_t peak, size_t n)
 		else if ( x[i] > x[lbound] && is_left_of_peak )
 		{
 			index_t cand = lbound;
-			index_t lwindow = (cand - 2) > 0 ? (cand - 2) : 0;
+			index_t lwindow = norm_ind(cand - 2, n);
 			i--;
 			// check if candidate is local minimum
 			while ( i >= lwindow )
@@ -1279,7 +1284,7 @@ index_t peak_rbound(T * x, index_t peak, size_t n)
 		else if ( x[i] > x[rbound] && is_right_of_peak )
 		{
 			index_t cand = rbound;
-			index_t rwindow = (cand + 2) < n - 1 ? (cand + 2) : n - 1;
+			index_t rwindow = norm_ind(cand + 2, n);
 			i++;
 			// check if candidate is local minimum
 			while ( i <= rwindow )
@@ -1712,11 +1717,11 @@ size_t do_approx1(Tout * ptr, Tx * xi, size_t ni, Tx * x, Ty * y,
 			}
 		}
 	}
-	Free(processed);
+	R_Free(processed);
 	if ( need_sort )
 	{
-		Free(xs);
-		Free(ys);
+		R_Free(xs);
+		R_Free(ys);
 	}
 	return num_matches;
 }
